/*------------------------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------*\
 * In dieser Datei werden die moeglichen Blinkmodi definiert bzw. erzeugt.                  *
\*------------------------------------------------------------------------------------------*/

#include <linux/version.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/avm_led.h>
#include <linux/avm_profile.h>
#include <linux/avm_debug.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#include <linux/kallsyms.h>
#endif
#include <asm/uaccess.h>
#if defined(CONFIG_MIPS)
#include <asm/mipsregs.h>
#endif
#if defined(CONFIG_ARM)
#endif
#include "avm_sammel.h"
#include "avm_led.h"

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#include <asm/mach_avm.h>
#else
#endif

/*------------------------------------------------------------------------------------------*\
 * avm_led_profile.c
\*------------------------------------------------------------------------------------------*/
int avm_led_profile_init(unsigned int table_index, struct _avm_virt_led *V) {
    switch(table_index) {
        case 0:  /* off */
            printk("[avm_led] profile_init: disable\n");
            return 1;
        case 1: /* on */
            printk("[avm_led] profile_init: enable\n");
            return 2;
        case 2: /* write CSV File */
            printk("[avm_led] profile_init: write CSV file\n");
            return 3;
        case 3: /* write ASCII File */
            printk("[avm_led] profile_init: write ASCII file\n");
            return 4;
        case 4: /* display ASCII*/
            printk("[avm_led] profile_init: display ASCII\n");
            return 5;
        case 5: /* on (wraparround-mode) */
            printk("[avm_led] profile_init: enable (wraparround-mode)\n");
            return 6;
        case 6: /* on skb trace */
            printk("[avm_led] profile_init: enable skb trace\n");
            return 7;
        case 7: /* on skb trace (wraparround-mode) */
            printk("[avm_led] profile_init: enable skb trace (wraparround-mode)\n");
            return 8;
        default:
            printk("[avm_led] profile_init: other %u\n", table_index);
            return 33;
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avm_led_profile_exit(int handle) {
    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avm_led_profile_stop(int handle) {
    return;
}

#define MIPS_PERFOMANCE_IRQ_ENABLE                  (1 << 4)
#define MIPS_PERFOMANCE_USER_MODE_ENABLE            (1 << 3)
#define MIPS_PERFOMANCE_SUPERVISOR_MODE_ENABLE      (1 << 2)
#define MIPS_PERFOMANCE_KERNEL_MODE_ENABLE          (1 << 1)
#define MIPS_PERFOMANCE_EXCEPTION_ENABLE            (1 << 0)

char *print_perfomance_counter_mode(unsigned int mode) {
    static char Buffer[128];
    sprintf(Buffer, "User-Mode %s Supervisor-Mode %s Kernel-Mode %s Exeption-Mode %s",
                            (MIPS_PERFOMANCE_USER_MODE_ENABLE & mode) ? "enabled" : "disable",
                            (MIPS_PERFOMANCE_SUPERVISOR_MODE_ENABLE & mode) ? "enabled" : "disable",
                            (MIPS_PERFOMANCE_KERNEL_MODE_ENABLE & mode) ? "enabled" : "disable",
                            (MIPS_PERFOMANCE_EXCEPTION_ENABLE & mode) ? "enabled" : "disable");
    return Buffer;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avm_led_profile_timer(unsigned long handle) {
    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
char *avm_profile_data_short_names[] = {
    "FREE", /*--- avm_profile_data_type_free, ---*/
    "TEXT", /*--- avm_profile_data_type_text, ---*/
    "CODE", /*--- avm_profile_data_type_code_address_info, ---*/
    "DATA", /*--- avm_profile_data_type_data_address_info, ---*/
    "SKB",  /*--- avm_profile_data_type_trace_skb, ---*/
    "BIRQ", /*--- avm_profile_data_type_hw_irq_begin, ---*/
    "EIRQ", /*--- avm_profile_data_type_hw_irq_end, ---*/
    "BSWI", /*--- avm_profile_data_type_sw_irq_begin, ---*/
    "ESWI", /*--- avm_profile_data_type_sw_irq_end, ---*/
    "BTIM", /*--- avm_profile_data_type_timer_begin, ---*/
    "ETIM", /*--- avm_profile_data_type_timer_end, ---*/
    "BLET", /*--- avm_profile_data_type_tasklet_begin, ---*/
    "ELET", /*--- avm_profile_data_type_tasklet_end, ---*/
    "BLHT", /*--- avm_profile_data_type_hi_tasklet_begin, ---*/
    "ELHT", /*--- avm_profile_data_type_hi_tasklet_end, ---*/
    "BWRK", /*--- avm_profile_data_type_workitem_begin, ---*/
    "EWRK", /*--- avm_profile_data_type_workitem_end, ---*/
    "BCPT", /*--- avm_profile_data_type_cpphytx_begin, ---*/
    "ECPT", /*--- avm_profile_data_type_cpphytx_end, ---*/
    "BCPR", /*--- avm_profile_data_type_cpphyrx_begin, ---*/
    "ECPR", /*--- avm_profile_data_type_cpphyrx_end, ---*/
    "BFUN", /*--- avm_profile_data_type_func_begin, ---*/
    "EFUN", /*--- avm_profile_data_type_func_end, ---*/
    "ERROR", /*--- avm_profile_data_type_unknown ---*/
    NULL
};

char *avm_profile_data_long_names[] = {
    "free", /*--- avm_profile_data_type_free, ---*/
    "text", /*--- avm_profile_data_type_text, ---*/
    "code", /*--- avm_profile_data_type_code_address_info, ---*/
    "data", /*--- avm_profile_data_type_data_address_info, ---*/
    "skb",  /*--- avm_profile_data_type_trace_skb, ---*/
    "begin hw irq", /*--- avm_profile_data_type_hw_irq_begin, ---*/
    "end hw irq", /*--- avm_profile_data_type_hw_irq_end, ---*/
    "begin sw irq", /*--- avm_profile_data_type_sw_irq_begin, ---*/
    "end sw irq", /*--- avm_profile_data_type_sw_irq_end, ---*/
    "begin timer", /*--- avm_profile_data_type_timer_begin, ---*/
    "end timer", /*--- avm_profile_data_type_timer_end, ---*/
    "begin tasklet", /*--- avm_profile_data_type_tasklet_begin, ---*/
    "end tasklet", /*--- avm_profile_data_type_tasklet_end, ---*/
    "begin hitasklet", /*--- avm_profile_data_type_hi_tasklet_begin, ---*/
    "end hitasklet", /*--- avm_profile_data_type_hi_tasklet_end, ---*/
    "begin workitem", /*--- avm_profile_data_type_workitem_begin, ---*/
    "end workitem", /*--- avm_profile_data_type_workitem_end, ---*/
    "begin cpphy_tx", /*--- avm_profile_data_type_cpphytx_begin, ---*/
    "end cpphy_tx", /*--- avm_profile_data_type_cpphytx_end, ---*/
    "begin cpphy_rx", /*--- avm_profile_data_type_cpphyrx_begin, ---*/
    "end cpphy_rx", /*--- avm_profile_data_type_cpphyrx_end, ---*/
    "begin func", /*--- avm_profile_data_type_func_begin, ---*/
    "end func", /*--- avm_profile_data_type_func_end, ---*/
    "unknown", /*--- avm_profile_data_type_unknown ---*/
    NULL
};

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#ifdef CONFIG_MIPS
char *perfomance_counter_options[2][64] = {
    {
        /*--- index  0 ---*/ "Cycles",
        /*--- index  1 ---*/ "Instructions completed",
        /*--- index  2 ---*/ "branch instructions",
        /*--- index  3 ---*/ "JR r31 (return) instructions",
        /*--- index  4 ---*/ "JR (not r31) instructions",
        /*--- index  5 ---*/ "ITLB accesses",
        /*--- index  6 ---*/ "DTLB accesses",
        /*--- index  7 ---*/ "JTLB instruction accesses",
        /*--- index  8 ---*/ "JTLB data accesses",
        /*--- index  9 ---*/ "Instruction Cache accesses",
        /*--- index 10 ---*/ "Data cache accesses",
        /*--- index 11 ---*/ "Data cache misses",
        /*--- index 12 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 13 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 14 ---*/ "integer instructions completed",
        /*--- index 15 ---*/ "loads completed",
        /*--- index 16 ---*/ "J/JAL completed",
        /*--- index 17 ---*/ "no-ops completed",
        /*--- index 18 ---*/ "Stalls",
        /*--- index 19 ---*/ "SC instructions completed",
        /*--- index 20 ---*/ "Prefetch instructions completed",
        /*--- index 21 ---*/ "L2 cache writebacks",
        /*--- index 22 ---*/ "L2 cache misses",
        /*--- index 23 ---*/ "Exceptions taken",
        /*--- index 24 ---*/ "cache ﬁxup",
        /*--- index 25 ---*/ "IFU stalls",
        /*--- index 26 ---*/ "DSP Instructions Completed",
        /*--- index 27 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 28 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 29 ---*/ "Impl. speciﬁc ISPRAM event",
        /*--- index 30 ---*/ "Impl. speciﬁc CorExtend event",
        /*--- index 31 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 32 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 33 ---*/ "Uncached Loads",
        /*--- index 34 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 35 ---*/ "CP2 Arithmetic Instns Completed",
        /*--- index 36 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 37 ---*/ "I$ Miss Stall cycles",
        /*--- index 38 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 39 ---*/ "D$ miss cycles",
        /*--- index 40 ---*/ "Uncached stall cycles",
        /*--- index 41 ---*/ "MDU stall cycles",
        /*--- index 42 ---*/ "CP2 stall cycles",
        /*--- index 43 ---*/ "ISPRAM Stall Cycles",
        /*--- index 44 ---*/ "CACHE Instn stall cycles",
        /*--- index 45 ---*/ "Load to Use stalls",
        /*--- index 46 ---*/ "Other interlock stalls",
        /*--- index 47 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 48 ---*/ "IFU FB full refetches",
        /*--- index 49 ---*/ "EJTAG Instruction Triggerpoints",
        /*--- index 50 ---*/ "FSB < 1/4 full",
        /*--- index 51 ---*/ "FSB > 1/2 full",
        /*--- index 52 ---*/ "LDQ < 1/4 full",
        /*--- index 53 ---*/ "LDQ > 1/2 full",
        /*--- index 54 ---*/ "WBB < 1/4 full",
        /*--- index 55 ---*/ "WBB > 1/2 full",
        /*--- index 56 ---*/ NULL, /*--- "Reserved", ---*/
        NULL
    },
    {
        /*--- index  0 ---*/ "Cycles",
        /*--- index  1 ---*/ "Instructions completed",
        /*--- index  2 ---*/ "Branch mispredictions",
        /*--- index  3 ---*/ "JR r31 mispredictions",
        /*--- index  4 ---*/ "JR r31 not predicted",
        /*--- index  5 ---*/ "ITLB misses",
        /*--- index  6 ---*/ "DTLB misses",
        /*--- index  7 ---*/ "JTLB instruction misses",
        /*--- index  8 ---*/ "JTLB data misses",
        /*--- index  9 ---*/ "Instruction cache misses",
        /*--- index 10 ---*/ "Data cache writebacks",
        /*--- index 11 ---*/ "Data cache misses",
        /*--- index 12 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 13 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 14 ---*/ "FPU instructions completed",
        /*--- index 15 ---*/ "stores completed",
        /*--- index 16 ---*/ "MIPS16 instructions completed",
        /*--- index 17 ---*/ "integer multiply/divide completed",
        /*--- index 18 ---*/ "replay traps (other than uTLB)",
        /*--- index 19 ---*/ "SC instructions failed",
        /*--- index 20 ---*/ "Prefetch instructions completed with cache hit",
        /*--- index 21 ---*/ "L2 cache accesses",
        /*--- index 22 ---*/ "L2 cache single bit errors corrected",
        /*--- index 23 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 24 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 25 ---*/ "ALU stalls",
        /*--- index 26 ---*/ "ALU-DSP Saturations Done",
        /*--- index 27 ---*/ "MDU-DSP Saturations Done",
        /*--- index 28 ---*/ "Impl. speciﬁc Cp2 event",
        /*--- index 29 ---*/ "Impl. speciﬁc DSPRAM event",
        /*--- index 30 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 31 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 32 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 33 ---*/ "Uncached Stores",
        /*--- index 34 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 35 ---*/ "CP2 To/From Instns completed",
        /*--- index 36 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 37 ---*/ "D$ miss stall cycles",
        /*--- index 38 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 39 ---*/ "L2 miss cycles",
        /*--- index 40 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 41 ---*/ "FPU stall cycles",
        /*--- index 42 ---*/ "CorExtend stall cycles",
        /*--- index 43 ---*/ "DSPRAM stall cycles",
        /*--- index 44 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 45 ---*/ "ALU to AGEN stalls",
        /*--- index 46 ---*/ "Branch mispredict stalls",
        /*--- index 47 ---*/ NULL, /*--- "Reserved", ---*/
        /*--- index 48 ---*/ "FB entry allocated",
        /*--- index 49 ---*/ "EJTAG Data Triggerpoints",
        /*--- index 50 ---*/ "FSB 1/4-1/2 full",
        /*--- index 51 ---*/ "FSB full pipeline stalls",
        /*--- index 52 ---*/ "LDQ 1/4-1/2 full",
        /*--- index 53 ---*/ "LDQ full pipeline stalls",
        /*--- index 54 ---*/ "WBB 1/4-1/2 full",
        /*--- index 55 ---*/ "WBB full pipeline stalls",
        /*--- index 56 ---*/ NULL, /*--- "Reserved" ---*/
        NULL
    }
};
#endif /*--- #ifdef CONFIG_MIPS ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avm_led_profile_action(int handle, unsigned int trace_mask, unsigned int param2) {
    /*--- extern struct _avm_profile_data **avm_simple_profiling_enable(unsigned int on, unsigned int *count, unsigned long *timediff); ---*/
    unsigned int len, i, ii = 0;
    unsigned long timediff;
    static unsigned char Buffer[512], Symbols[256];
    static struct file *fp;
    mm_segment_t  oldfs;  
    size_t bytesWritten;
    char *p = Buffer;

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    void open_output_file(char *name) {
        fp = filp_open(name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if(IS_ERR(fp)) {
            printk(KERN_ERR "[avm_led] Failed: Could not open \"/var/profile.log\"\n");
            return;
        }
        /* Lesezugriff auf File(system) erlaubt? */
        if (fp->f_op->write == NULL) {
            printk(KERN_ERR "[avm_led] Failed: Could not write \"/var/profile.log\"\n");
            return;
        }

        /* Von Anfang an lesen */
        fp->f_pos = 0;
    }

    void close_output_file(void) {
        filp_close(fp, NULL);
    }

    printk("[avm_led_profile_action] handle %u\n", handle);

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    switch(handle) {
        default:
            break;
        case 1: /* disable */
            avm_simple_profiling_enable(0, 0, &len, &timediff);
            printk("[avm_led] profile_action: disable, %u entries recorded (output: 2: CSV, 3: ASCII)\n", len);
            return;
        case 2: /* enable */
        case 6: /* enable wraparound */
        case 7: /* enable skb trace */
        case 8: /* enable skb trace wraparound */
            avm_simple_profiling_enable(handle, trace_mask, &len, NULL);
            printk("[avm_led] profile_action: enable\n");
            return;
        case 3: /* write CSV file */
            avm_simple_profiling_enable(0, 0, &len, &timediff);
            printk("[avm_led] profile_action: %u entries recorded, write CSV file\n", len);
            open_output_file("/var/profile.csv");
            sprintf(Buffer, "# measure time %lu msec\n", (timediff * 1000) / HZ);

#ifdef CONFIG_FUSIV_VX180
            i = read_c0_perfctrl0();
            /*--- printk(KERN_ERR "read_c0_perfctrl0: 0x%x\n", i); ---*/
            sprintf(Buffer + strlen(Buffer), "# performance 0 \"%s\" (%s)\n", 
                    perfomance_counter_options[0][(i >> 5) & ((1 << 7) - 1)],
                    print_perfomance_counter_mode(i));

            i = read_c0_perfctrl1();
            /*--- printk(KERN_ERR "read_c0_perfctrl1: 0x%x\n", i); ---*/
            sprintf(Buffer + strlen(Buffer), "# performance 1 \"%s\" (%s)\n", 
                    perfomance_counter_options[1][(i >> 5) & ((1 << 7) - 1)],
                    print_perfomance_counter_mode(i));
#endif /*--- #ifdef CONFIG_FUSIV_VX180 ---*/

            p = &Buffer[strlen(Buffer)];
            goto write_file;
            return;
        case 4: /* write ASCII file */
            avm_simple_profiling_enable(0, 0, &len, &timediff);
            printk("[avm_led] profile_action: %u entries recorded, write ASCII file\n", len);
            open_output_file("/var/profile.txt");
            sprintf(Buffer, "# measure time %lu msec\n", (timediff * 1000) / HZ);

#ifdef CONFIG_FUSIV_VX180
            i = read_c0_perfctrl0();
            sprintf(Buffer + strlen(Buffer), "# performance 0 %s (%s)\n", 
                    perfomance_counter_options[0][(i >> 5) & ((1 << 7) - 1)],
                    print_perfomance_counter_mode(i));

            i = read_c0_perfctrl1();
            sprintf(Buffer + strlen(Buffer), "# performance 1 %s (%s)\n", 
                    perfomance_counter_options[1][(i >> 5) & ((1 << 7) - 1)],
                    print_perfomance_counter_mode(i));
#endif /*--- #ifdef CONFIG_FUSIV_VX180 ---*/

            p = &Buffer[strlen(Buffer)];
write_file:

            /*------------------------------------------------------------------------------*\
            \*------------------------------------------------------------------------------*/
            for(i = 0 ; i < len ; i++) {
                unsigned int _addr;
                unsigned int len;
                char *_text;
                char *description = "";
                struct _avm_profile_data *data;

                data = avm_simple_profiling_by_idx(i);
                if(data == NULL) {
                    break;
                }
                /*--------------------------------------------------------------------------*\
                \*------------------------------------------------------------------------------*/
                if((data->type < avm_profile_data_type_free) || (data->type > avm_profile_data_type_func_end)) {
                    printk("[profiling:%d] internal error data type %d unknown\n", i, data->type);
                    continue;
                }

                /*--------------------------------------------------------------------------*\
                \*------------------------------------------------------------------------------*/
                /*--- if((data->curr < (struct task_struct *)0x94000000) || (data->curr > (struct task_struct *)0x96000000)) { ---*/
                    /*--- if((data->curr < (struct task_struct *)0xC0000000) || (data->curr > (struct task_struct *)0xE0000000)) { ---*/
                        /*--- printk("[profiling:%d %s] invallid current pointer 0x%x\n", i, ---*/ 
                                /*--- avm_profile_data_short_names[data->type], ---*/ 
                                /*--- (void *)data->curr); ---*/
                        /*--- continue; ---*/
                    /*--- } ---*/
                /*--- } ---*/

                /*--------------------------------------------------------------------------*\
                \*------------------------------------------------------------------------------*/
                /*--- if((data->addr < 0x94000000) || (data->addr > 0x96000000)) { ---*/
                    /*--- printk("[profiling:%d %s] invallid addr pointer 0x%x\n", i, ---*/ 
                            /*--- avm_profile_data_short_names[data->type], ---*/ 
                            /*--- data->addr); ---*/
                    /*--- continue; ---*/
                /*--- } ---*/

                /*--------------------------------------------------------------------------*\
                \*------------------------------------------------------------------------------*/
                switch(data->type) {
                    case avm_profile_data_type_free:
                    case avm_profile_data_type_unknown:
                    default:
                        /*--- printk("x"); ---*/
                        break;
                    case avm_profile_data_type_code_address_info:
                        sprint_symbol(Symbols, data->addr);
                        switch(handle) {
                            case 3:
                                sprintf(p, "0x%08X;0x%08X;0x%08X;CODE;0x%08x;%s;%.*s;%u\n", 
                                        data->time, 
                                        data->total_access, 
                                        data->total_activate,
                                        data->addr, Symbols, TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                printk("c");
                                break;
                            case 4:
                                sprintf(p, "0x%08X: CODE 0x%08x %s (%.*s interrupted by irq %u)\n", 
                                        data->time, 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                printk("c");
                                break;
                            case 5:
                                printk("0x%08X: CODE 0x%08x %s (%.*s interrupted by irq %u)\n", 
                                        data->time, 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                break;
                        }
                        break;
                    case avm_profile_data_type_data_address_info:
                        sprint_symbol(Symbols, data->addr);
                        switch(handle) {
                            case 3:
                                sprintf(p, "0x%08X;0x%08X;0x%08X;DATA;0x%08x;%s;%.*s;%u\n", 
                                        data->total_access, 
                                        data->total_activate, 
                                        data->time, 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                printk("d");
                                break;
                            case 4:
                                sprintf(p, "0x%08X: DATA 0x%08x %s (%.*s interrupted by irq %u)\n", 
                                        data->time, 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                printk("d");
                                break;
                            case 5:
                                printk("0x%08X: DATA 0x%08x %s (%.*s interrupted by irq %u)\n", 
                                        data->time, 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                break;
                        }
                        break;
                    case avm_profile_data_type_trace_skb:
                        printk("s");
                        goto print_work_trace;
                    case avm_profile_data_type_text:
                        /*--- printk("t"); ---*/
                        break;
                    case avm_profile_data_type_hw_irq_begin:
                    case avm_profile_data_type_hw_irq_end:
                        description = "interrupted by irq";
                        goto print_work_trace;
                    case avm_profile_data_type_sw_irq_begin:
                    case avm_profile_data_type_sw_irq_end:
                        description = "id:";
                        goto print_work_trace;
                    case avm_profile_data_type_timer_begin:
                    case avm_profile_data_type_timer_end:
                        description = "id:";
                        goto print_work_trace;
                    case avm_profile_data_type_tasklet_begin:
                    case avm_profile_data_type_tasklet_end:
                        description = "id:";
                        goto print_work_trace;
                    case avm_profile_data_type_hi_tasklet_begin:
                    case avm_profile_data_type_hi_tasklet_end:
                        description = "id:";
                        goto print_work_trace;
                    case avm_profile_data_type_workitem_begin:
                    case avm_profile_data_type_workitem_end:
                        description = "id:";
                        goto print_work_trace;
                    case avm_profile_data_type_cpphyrx_begin:
                    case avm_profile_data_type_cpphyrx_end:
                    case avm_profile_data_type_cpphytx_begin:
                    case avm_profile_data_type_cpphytx_end:
                    case avm_profile_data_type_func_begin:
                    case avm_profile_data_type_func_end:
                        description = "id:";
                        goto print_work_trace;

print_work_trace:
                        sprint_symbol(Symbols, data->addr);
                        switch(handle) {
                            case 3:
                                sprintf(p, "0x%08X;0x%08X;0x%08X;%s;0x%08x;%s;%.*s;%u;\n", 
                                        data->time, 
                                        data->total_access, 
                                        data->total_activate, 
                                        avm_profile_data_short_names[data->type], 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, 
                                        data->id);
                                printk("d");
                                break;
                            case 4:
                                sprintf(p, "0x%08X:%s 0x%08x %s (%.*s %s %u);\n", 
                                        data->time, 
                                        avm_profile_data_short_names[data->type], 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, description, 
                                        data->id);
                                printk("d");
                                break;
                            case 5:
                                printk("0x%08X: %s 0x%08x %s (%.*s %s %u)\n", 
                                        data->time, 
                                        avm_profile_data_long_names[data->type], 
                                        data->addr, 
                                        Symbols, 
                                        TASK_COMM_LEN, 
                                        data->curr->comm, description, 
                                        data->id);
                                break;
                        }
                }
                if(handle != 5) {
                    len = strlen(Buffer);
                    oldfs = get_fs();
                    set_fs(KERNEL_DS);
                    bytesWritten = fp->f_op->write(fp, Buffer, len, &fp->f_pos);
                    if(bytesWritten != len) {
                        printk(KERN_ERR "memory full (abort writing file)\n");
                        set_fs(oldfs);
                        break;
                    }
                    p = Buffer;
                    set_fs(oldfs);
                }
            }
            if(handle != 5) {
                close_output_file();
                printk("\n[avm_led] file closed\n");
            }
            return;
    }
    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avm_led_profile_sync(int virt_led_handle) {
    return;
}

#if defined(CONFIG_AVM_DEBUG) || defined(CONFIG_AVM_DEBUG_MODULE)
static void *profiler_DbgHandle;
#define SKIP_SPACE(a)       while(*(a) && ((*(a) == ' ') || (*(a) == '\t'))) (a)++ 
#define SKIP_NON_SPACE(a)   while(*(a) && ((*(a) != ' ') && (*(a) != '\t'))) (a)++ 
/*--------------------------------------------------------------------------------*\
\*--------------------------------------------------------------------------------*/
static void cmdlineparse(char *cmdline, void *dummy) {
    char *p = cmdline; 
    unsigned int  val;
    SKIP_SPACE(p);
    if((!*p || !strcmp(p, "help"))) {
        printk(KERN_ERR "use: echo profiler <cmd-nr> [mask=0x<trace-mask>] [count=<sample-count/%d>]\n", profile_DataSetsPerBlock);
        printk(KERN_ERR "\t0: disable, 1: enable,2: write csv,3: write ascii,4: display ascii,5: enable+wraparound,6: enable skb,7: skb+wraparound\n");
        printk(KERN_ERR "use: echo profiler performance [<register-nr>] [<Mode-nr>|<Mode-name>|liste]\n");
        return;
    }
    if((*p >= '0') && (*p <= '9')) {
        unsigned int trace_mask = 
            /*--- (1 << avm_profile_data_type_text) | ---*/
            (1 << avm_profile_data_type_code_address_info) |
            (1 << avm_profile_data_type_data_address_info) |
            /*--- (1 << avm_profile_data_type_trace_skb) | ---*/
            (1 << avm_profile_data_type_hw_irq_begin) |
            (1 << avm_profile_data_type_hw_irq_end) |
            (1 << avm_profile_data_type_sw_irq_begin) |
            (1 << avm_profile_data_type_sw_irq_end) |
            (1 << avm_profile_data_type_timer_begin) |
            (1 << avm_profile_data_type_timer_end) |
            (1 << avm_profile_data_type_tasklet_begin) |
            (1 << avm_profile_data_type_tasklet_end) |
            (1 << avm_profile_data_type_hi_tasklet_begin) |
            (1 << avm_profile_data_type_hi_tasklet_end) |
            (1 << avm_profile_data_type_workitem_begin) |
            (1 << avm_profile_data_type_workitem_end) |
            (1 << avm_profile_data_type_cpphytx_begin) |
            (1 << avm_profile_data_type_cpphytx_end) |
            (1 << avm_profile_data_type_cpphyrx_begin) |
            (1 << avm_profile_data_type_cpphyrx_end) |
            /*--- (1 << avm_profile_data_type_func_begin) | ---*/
            /*--- (1 << avm_profile_data_type_func_end) | ---*/
            0;
        if(strstr(p, "count=")) {
            sscanf(strstr(p, "count="), "count=%d", &profile_BlockNeeded);
        }
        if(strstr(p, "mask=0x")) {
            sscanf(strstr(p, "mask=0x"), "mask=0x%x", &trace_mask);
        }
        sscanf(p, "%x", &val);
        avm_led_profile_action(val+1, trace_mask, 0);
        return;
    }
#ifdef CONFIG_MIPS
    if(!strncmp(p, "perform", sizeof("perform") - 1)) {
        int i;
        int perf_reg_number = 0;
        int perf_reg_flags = 0;
        int perf_reg_index = 0;
        int perf_max_registers = 0;

        i = read_c0_config1();
        if((i & (1 << 4)) == 0) {
            printk(KERN_ERR "[simple-profiling]: no performance counters implemented\n");
            return;
        }
        perf_max_registers = 1;
        i = read_c0_perfctrl0();
        if(i & (1U << 31)) {
            perf_max_registers = 2;
            i = read_c0_perfctrl1();
            if(i & (1U << 31)) {
                perf_max_registers = 3;
                i = read_c0_perfctrl2();
                if(i & (1U << 31)) {
                    perf_max_registers = 4;
                    i = read_c0_perfctrl3();
                    if(i & (1U << 31)) {
                        perf_max_registers = 5;
                    }
                }
            }
        }
        printk(KERN_ERR "[simple-profiling]: %d performance counters implemented\n", perf_max_registers);

        SKIP_NON_SPACE(p);
        SKIP_SPACE(p);
        if((*p >= '0') && (*p <= '3')) {
            sscanf(p, "%d", &perf_reg_number);
        }
        SKIP_NON_SPACE(p);
        SKIP_SPACE(p);
        if(!*p) {
show_performance_counter_options:
            printk(KERN_INFO "Optionen des Performance Counter %d:\n", i);
            for(i = 0 ; i < 64 ; i++) {
                if(perfomance_counter_options[perf_reg_number][i] == NULL)
                    continue;
                printk(KERN_INFO "\t[%2d]: %s\n", i, perfomance_counter_options[perf_reg_number][i]);
            }
            return;
        }
        if((*p >= '0') && (*p <= '9')) {
            sscanf(p, "%d", &perf_reg_index);
        } else {
            for(i = 0 ; i < 64 ; i++) {
                if(perfomance_counter_options[perf_reg_number][i] == NULL)
                    continue;
                if(!strncmp(perfomance_counter_options[perf_reg_number][i], p , strlen(p))) {
                    perf_reg_index = i;
                    break;
                }
            }
            if(i == 64) {
                goto show_performance_counter_options;
            }
        }
        SKIP_NON_SPACE(p);
        SKIP_SPACE(p);

        perf_reg_flags = 
                /*--- MIPS_PERFOMANCE_IRQ_ENABLE | ---*/
                MIPS_PERFOMANCE_USER_MODE_ENABLE |
                MIPS_PERFOMANCE_SUPERVISOR_MODE_ENABLE |
                MIPS_PERFOMANCE_KERNEL_MODE_ENABLE |
                /*--- MIPS_PERFOMANCE_EXCEPTION_ENABLE | ---*/
                0;

        printk(KERN_INFO "Enable Performance Counter %d for %s (%s)\n",
            perf_reg_number, perfomance_counter_options[perf_reg_number][perf_reg_index ],
            print_perfomance_counter_mode(perf_reg_flags));

        i = (perf_reg_index << 5) | perf_reg_flags;
        switch(perf_reg_number) {
            case 0: write_c0_perfctrl0(i); break;
            case 1: write_c0_perfctrl1(i); break;
            case 2: write_c0_perfctrl2(i); break;
            case 3: write_c0_perfctrl3(i); break;
            default:
                    break;
        }


/*--- #define read_c0_perfctrl0()	__read_32bit_c0_register($25, 0) ---*/
/*--- #define read_c0_perfcntr0()	__read_32bit_c0_register($25, 1) ---*/
/*--- #define write_c0_perfcntr0(val)	__write_32bit_c0_register($25, 1, val) ---*/
/*--- #define read_c0_perfctrl1()	__read_32bit_c0_register($25, 2) ---*/
/*--- #define read_c0_perfcntr1()	__read_32bit_c0_register($25, 3) ---*/
/*--- #define write_c0_perfcntr1(val)	__write_32bit_c0_register($25, 3, val) ---*/
/*--- #define read_c0_perfctrl2()	__read_32bit_c0_register($25, 4) ---*/
/*--- #define write_c0_perfctrl2(val)	__write_32bit_c0_register($25, 4, val) ---*/
/*--- #define read_c0_perfcntr2()	__read_32bit_c0_register($25, 5) ---*/
/*--- #define write_c0_perfcntr2(val)	__write_32bit_c0_register($25, 5, val) ---*/
/*--- #define read_c0_perfctrl3()	__read_32bit_c0_register($25, 6) ---*/
/*--- #define write_c0_perfctrl3(val)	__write_32bit_c0_register($25, 6, val) ---*/
/*--- #define read_c0_perfcntr3()	__read_32bit_c0_register($25, 7) ---*/
/*--- #define write_c0_perfcntr3(val)	__write_32bit_c0_register($25, 7, val) ---*/

    }
#endif /*--- #ifdef CONFIG_MIPS ---*/
}
/*--------------------------------------------------------------------------------*\
 * Das Led-Modul ist komplett ausgelagert worden, so ist es nun notwendig
 * den Profiler ebenfalls anders anzusteuern:
 * echo profiler <val> >/dev/debug
\*--------------------------------------------------------------------------------*/
void __init avm_profiler_init(void) {
    profiler_DbgHandle = avm_DebugCallRegister("profiler", cmdlineparse, NULL);
#ifdef CONFIG_MIPS
    cmdlineparse("perform 0 10", "");
    cmdlineparse("perform 1 11", "");
#endif /*--- #ifdef CONFIG_MIPS ---*/
}
#if defined(CONFIG_AVM_DEBUG_MODULE) && (CONFIG_AVM_DEBUG_MODULE == 1)
/*--------------------------------------------------------------------------------*\
\*--------------------------------------------------------------------------------*/
void avm_profiler_exit(void) {
    if(profiler_DbgHandle) {
        avm_DebugCallUnRegister(profiler_DbgHandle);
        profiler_DbgHandle = NULL;
    }
}
#endif/*--- #if defined(CONFIG_AVM_DEBUG_MODULE) && (CONFIG_AVM_DEBUG_MODULE == 1) ---*/
#endif/*--- #if defined(CONFIG_AVM_DEBUG) || defined(CONFIG_AVM_DEBUG_MODULE) ---*/
